@@ -61,7 +61,7 @@ else:
 
 
 __author__ = 'Alec Thomas <alec@swapoff.org>'
-__version__ = '0.20.1'
+__version__ = '0.21.0'
 __version_tag__ = ''
 
 log = logging.getLogger('injector')
@@ -254,7 +254,7 @@ class Provider(Generic[T]):
         raise NotImplementedError  # pragma: no cover
 
 
-class ClassProvider(Provider):
+class ClassProvider(Provider, Generic[T]):
     """Provides instances from a given class, created using an Injector."""
 
     def __init__(self, cls: Type[T]) -> None:
@@ -264,7 +264,7 @@ class ClassProvider(Provider):
         return injector.create_object(self._cls)
 
 
-class CallableProvider(Provider):
+class CallableProvider(Provider, Generic[T]):
     """Provides something using a callable.
 
     The callable is called every time new value is requested from the provider.
@@ -305,7 +305,7 @@ class CallableProvider(Provider):
         return '%s(%r)' % (type(self).__name__, self._callable)
 
 
-class InstanceProvider(Provider):
+class InstanceProvider(Provider, Generic[T]):
     """Provide a specific instance.
 
     ::
@@ -379,6 +379,13 @@ class Binding(_BindingBase):
         return _get_origin(_punch_through_alias(self.interface)) in {dict, list}
 
 
+@private
+class ImplicitBinding(Binding):
+    """A binding that was created implicitly by auto-binding."""
+
+    pass
+
+
 _InstallableModuleType = Union[Callable[['Binder'], None], 'Module', Type['Module']]
 
 
@@ -392,7 +399,9 @@ class Binder:
     _bindings: Dict[type, Binding]
 
     @private
-    def __init__(self, injector: 'Injector', auto_bind: bool = True, parent: 'Binder' = None) -> None:
+    def __init__(
+        self, injector: 'Injector', auto_bind: bool = True, parent: Optional['Binder'] = None
+    ) -> None:
         """Create a new Binder.
 
         :param injector: Injector we are binding for.
@@ -460,7 +469,7 @@ class Binder:
         self,
         interface: Type[List[T]],
         to: Union[List[T], Callable[..., List[T]], Provider[List[T]]],
-        scope: Union[Type['Scope'], 'ScopeDecorator'] = None,
+        scope: Union[Type['Scope'], 'ScopeDecorator', None] = None,
     ) -> None:  # pragma: no cover
         pass
 
@@ -469,12 +478,12 @@ class Binder:
         self,
         interface: Type[Dict[K, V]],
         to: Union[Dict[K, V], Callable[..., Dict[K, V]], Provider[Dict[K, V]]],
-        scope: Union[Type['Scope'], 'ScopeDecorator'] = None,
+        scope: Union[Type['Scope'], 'ScopeDecorator', None] = None,
     ) -> None:  # pragma: no cover
         pass
 
     def multibind(
-        self, interface: type, to: Any, scope: Union['ScopeDecorator', Type['Scope']] = None
+        self, interface: type, to: Any, scope: Union['ScopeDecorator', Type['Scope'], None] = None
     ) -> None:
         """Creates or extends a multi-binding.
 
@@ -555,7 +564,7 @@ class Binder:
         instance(self)
 
     def create_binding(
-        self, interface: type, to: Any = None, scope: Union['ScopeDecorator', Type['Scope']] = None
+        self, interface: type, to: Any = None, scope: Union['ScopeDecorator', Type['Scope'], None] = None
     ) -> Binding:
         provider = self.provider_for(interface, to)
         scope = scope or getattr(to or interface, '__scope__', NoScope)
@@ -643,12 +652,18 @@ class Binder:
             # The special interface is added here so that requesting a special
             # interface with auto_bind disabled works
             if self._auto_bind or self._is_special_interface(interface):
-                binding = self.create_binding(interface)
+                binding = ImplicitBinding(*self.create_binding(interface))
                 self._bindings[interface] = binding
                 return binding, self
 
         raise UnsatisfiedRequirement(None, interface)
 
+    def has_binding_for(self, interface: type) -> bool:
+        return interface in self._bindings
+
+    def has_explicit_binding_for(self, interface: type) -> bool:
+        return self.has_binding_for(interface) and not isinstance(self._bindings[interface], ImplicitBinding)
+
     def _is_special_interface(self, interface: type) -> bool:
         # "Special" interfaces are ones that you cannot bind yourself but
         # you can request them (for example you cannot bind ProviderOf(SomeClass)
@@ -782,10 +797,25 @@ class SingletonScope(Scope):
         try:
             return self._context[key]
         except KeyError:
-            provider = InstanceProvider(provider.get(self.injector))
+            instance = self._get_instance(key, provider, self.injector)
+            provider = InstanceProvider(instance)
             self._context[key] = provider
             return provider
 
+    def _get_instance(self, key: Type[T], provider: Provider[T], injector: 'Injector') -> T:
+        if injector.parent and not injector.binder.has_explicit_binding_for(key):
+            try:
+                return self._get_instance_from_parent(key, provider, injector.parent)
+            except (CallError, UnsatisfiedRequirement):
+                pass
+        return provider.get(injector)
+
+    def _get_instance_from_parent(self, key: Type[T], provider: Provider[T], parent: 'Injector') -> T:
+        singleton_scope_binding, _ = parent.binder.get_binding(type(self))
+        singleton_scope = singleton_scope_binding.provider.get(parent)
+        provider = singleton_scope.get(key, provider)
+        return provider.get(parent)
+
 
 singleton = ScopeDecorator(SingletonScope)
 
@@ -829,7 +859,7 @@ class Module:
                             % (function.__name__, type(self), e)
                         ) from e
                     return_type = annotations['return']
-                    binding = function.__func__.__binding__ = Binding(
+                    binding = cast(Any, function.__func__).__binding__ = Binding(
                         interface=return_type, provider=binding.provider, scope=binding.scope
                     )
                 bind_method = binder.multibind if binding.is_multibinding() else binder.bind
@@ -864,9 +894,9 @@ class Injector:
 
     def __init__(
         self,
-        modules: Union[_InstallableModuleType, Iterable[_InstallableModuleType]] = None,
+        modules: Union[_InstallableModuleType, Iterable[_InstallableModuleType], None] = None,
         auto_bind: bool = True,
-        parent: 'Injector' = None,
+        parent: Optional['Injector'] = None,
     ) -> None:
         # Stack of keys currently being injected. Used to detect circular
         # dependencies.
@@ -896,7 +926,8 @@ class Injector:
     def _log_prefix(self) -> str:
         return '>' * (len(self._stack) + 1) + ' '
 
-    def get(self, interface: Type[T], scope: Union[ScopeDecorator, Type[Scope]] = None) -> T:
+    @synchronized(lock)
+    def get(self, interface: Type[T], scope: Union[ScopeDecorator, Type[Scope], None] = None) -> T:
         """Get an instance of the given interface.
 
         .. note::
@@ -940,7 +971,8 @@ class Injector:
         log.debug(
             '%sInjector.get(%r, scope=%r) using %r', self._log_prefix, interface, scope, binding.provider
         )
-        result = scope_instance.get(interface, binding.provider).get(self)
+        provider_instance = scope_instance.get(interface, binding.provider)
+        result = provider_instance.get(self)
         log.debug('%s -> %r', self._log_prefix, result)
         return result
 
